"""SubMaker module is used to generate subtitles from WordBoundary and SentenceBoundary events."""

from datetime import timedelta
from typing import List, Optional

from .srt_composer import Subtitle, compose
from .typing import TTSChunk


class SubMaker:
    """
    SubMaker is used to generate subtitles from WordBoundary and SentenceBoundary messages.
    """

    def __init__(self) -> None:
        self.cues: List[Subtitle] = []
        self.type: Optional[str] = None

    def feed(self, msg: TTSChunk) -> None:
        """
        Feed a WordBoundary or SentenceBoundary message to the SubMaker object.

        Args:
            msg (dict): The WordBoundary or SentenceBoundary message.

        Returns:
            None
        """
        if msg["type"] not in ("WordBoundary", "SentenceBoundary"):
            raise ValueError(
                "Invalid message type, expected 'WordBoundary' or 'SentenceBoundary'."
            )

        if self.type is None:
            self.type = msg["type"]
        elif self.type != msg["type"]:
            raise ValueError(
                f"Expected message type '{self.type}', but got '{msg['type']}'."
            )

        self.cues.append(
            Subtitle(
                index=len(self.cues) + 1,
                start=timedelta(microseconds=msg["offset"] / 10),
                end=timedelta(microseconds=(msg["offset"] + msg["duration"]) / 10),
                content=msg["text"],
            )
        )

    def get_srt(self) -> str:
        """
        Get the SRT formatted subtitles from the SubMaker object.

        Returns:
            str: The SRT formatted subtitles.
        """
        return compose(self.cues)

    def __str__(self) -> str:
        return self.get_srt()



"""The edge_playback package wraps the functionality of mpv and edge-tts to generate
text-to-speech (TTS) using edge-tts and then plays back the generated audio using mpv.
"""

from .__main__ import _main

__all__ = ["_main"]



"""This module contains functions to list all available voices and a class to find the
correct voice based on their attributes."""

import json
import ssl
from typing import Any, List, Optional

import aiohttp
import certifi
from typing_extensions import Unpack

from .constants import SEC_MS_GEC_VERSION, VOICE_HEADERS, VOICE_LIST
from .drm import DRM
from .typing import Voice, VoicesManagerFind, VoicesManagerVoice

_SSL_CTX = ssl.create_default_context(cafile=certifi.where())


async def __list_voices(
    session: aiohttp.ClientSession, ssl_ctx: ssl.SSLContext, proxy: Optional[str]
) -> List[Voice]:
    """
    Private function that makes the request to the voice list URL and parses the
    JSON response. This function is used by list_voices() and makes it easier to
    handle client response errors related to clock skew.

    Args:
        session (aiohttp.ClientSession): The aiohttp session to use for the request.
        ssl_ctx (ssl.SSLContext): The SSL context to use for the request.
        proxy (Optional[str]): The proxy to use for the request.

    Returns:
        List[Voice]: A list of voices and their attributes.
    """
    async with session.get(
        f"{VOICE_LIST}&Sec-MS-GEC={DRM.generate_sec_ms_gec()}"
        f"&Sec-MS-GEC-Version={SEC_MS_GEC_VERSION}",
        headers=DRM.headers_with_muid(VOICE_HEADERS),
        proxy=proxy,
        ssl=ssl_ctx,
        raise_for_status=True,
    ) as url:
        data: List[Any] = json.loads(await url.text())

    for voice in data:
        if "VoiceTag" not in voice:
            voice["VoiceTag"] = {}

        if "ContentCategories" not in voice["VoiceTag"]:
            voice["VoiceTag"]["ContentCategories"] = []

        if "VoicePersonalities" not in voice["VoiceTag"]:
            voice["VoiceTag"]["VoicePersonalities"] = []

    return data


async def list_voices(
    *, connector: Optional[aiohttp.BaseConnector] = None, proxy: Optional[str] = None
) -> List[Voice]:
    """
    List all available voices and their attributes.

    This pulls data from the URL used by Microsoft Edge to return a list of
    all available voices.

    Args:
        connector (Optional[aiohttp.BaseConnector]): The connector to use for the request.
        proxy (Optional[str]): The proxy to use for the request.

    Returns:
        List[Voice]: A list of voices and their attributes.
    """
    async with aiohttp.ClientSession(connector=connector, trust_env=True) as session:
        try:
            data = await __list_voices(session, _SSL_CTX, proxy)
        except aiohttp.ClientResponseError as e:
            if e.status != 403:
                raise

            DRM.handle_client_response_error(e)
            data = await __list_voices(session, _SSL_CTX, proxy)
    return data


class VoicesManager:
    """
    A class to find the correct voice based on their attributes.
    """

    def __init__(self) -> None:
        self.voices: List[VoicesManagerVoice] = []
        self.called_create: bool = False

    @classmethod
    async def create(
        cls, custom_voices: Optional[List[Voice]] = None
    ) -> "VoicesManager":
        """
        Creates a VoicesManager object and populates it with all available voices.
        """
        self = VoicesManager()
        voices = await list_voices() if custom_voices is None else custom_voices
        self.voices = [
            {**voice, "Language": voice["Locale"].split("-")[0]} for voice in voices
        ]
        self.called_create = True
        return self

    def find(self, **kwargs: Unpack[VoicesManagerFind]) -> List[VoicesManagerVoice]:
        """
        Finds all matching voices based on the provided attributes.
        """
        if not self.called_create:
            raise RuntimeError(
                "VoicesManager.find() called before VoicesManager.create()"
            )

        matching_voices = [
            voice for voice in self.voices if kwargs.items() <= voice.items()
        ]
        return matching_voices



"""Custom exceptions for the edge-tts package."""


class EdgeTTSException(Exception):
    """Base exception for the edge-tts package."""


class UnknownResponse(EdgeTTSException):
    """Raised when an unknown response is received from the server."""


class UnexpectedResponse(EdgeTTSException):
    """Raised when an unexpected response is received from the server.

    This hasn't happened yet, but it's possible that the server will
    change its response format in the future."""


class NoAudioReceived(EdgeTTSException):
    """Raised when no audio is received from the server."""


class WebSocketError(EdgeTTSException):
    """Raised when a WebSocket error occurs."""


class SkewAdjustmentError(EdgeTTSException):
    """Raised when an error occurs while adjusting the clock skew."""



"""Data models for edge-tts."""

# pylint: disable=too-few-public-methods

import argparse
import re
from dataclasses import dataclass

from typing_extensions import Literal


@dataclass
class TTSConfig:
    """
    Represents the internal TTS configuration for edge-tts's Communicate class.
    """

    voice: str
    rate: str
    volume: str
    pitch: str
    boundary: Literal["WordBoundary", "SentenceBoundary"]

    @staticmethod
    def validate_string_param(param_name: str, param_value: str, pattern: str) -> str:
        """
        Validates the given string parameter based on type and pattern.

        Args:
            param_name (str): The name of the parameter.
            param_value (str): The value of the parameter.
            pattern (str): The pattern to validate the parameter against.

        Returns:
            str: The validated parameter.
        """
        if not isinstance(param_value, str):
            raise TypeError(f"{param_name} must be str")
        if re.match(pattern, param_value) is None:
            raise ValueError(f"Invalid {param_name} '{param_value}'.")
        return param_value

    def __post_init__(self) -> None:
        """
        Validates the TTSConfig object after initialization.
        """

        # Possible values for voice are:
        # - Microsoft Server Speech Text to Speech Voice (cy-GB, NiaNeural)
        # - cy-GB-NiaNeural
        # - fil-PH-AngeloNeural
        # Always send the first variant as that is what Microsoft Edge does.
        if not isinstance(self.voice, str):
            raise TypeError("voice must be str")
        match = re.match(r"^([a-z]{2,})-([A-Z]{2,})-(.+Neural)$", self.voice)
        if match is not None:
            lang = match.group(1)
            region = match.group(2)
            name = match.group(3)
            if name.find("-") != -1:
                region = f"{region}-{name[:name.find('-')]}"
                name = name[name.find("-") + 1 :]
            self.voice = (
                "Microsoft Server Speech Text to Speech Voice"
                + f" ({lang}-{region}, {name})"
            )

        # Validate the rate, volume, and pitch parameters.
        self.validate_string_param(
            "voice",
            self.voice,
            r"^Microsoft Server Speech Text to Speech Voice \(.+,.+\)$",
        )
        self.validate_string_param("rate", self.rate, r"^[+-]\d+%$")
        self.validate_string_param("volume", self.volume, r"^[+-]\d+%$")
        self.validate_string_param("pitch", self.pitch, r"^[+-]\d+Hz$")


class UtilArgs(argparse.Namespace):
    """CLI arguments."""

    text: str
    file: str
    voice: str
    list_voices: bool
    rate: str
    volume: str
    pitch: str
    write_media: str
    write_subtitles: str
    proxy: str



"""A tiny library for composing SRT files.

Based on https://github.com/cdown/srt with parsing, subtitle modifying
functionality and Python 2 support removed. This is because of
https://github.com/rany2/edge-tts/issues/383.

Typing support was added, and more Python 3 features were used.

Copyright (c) 2014-2023 Christopher Down
Copyright (c) 2025- rany <rany@riseup.net>

This file is licensed under the MIT License (MIT).
See the LICENSE-MIT file for details.
"""

import functools
import logging
import re
from datetime import timedelta
from typing import Generator, List, Union

LOG = logging.getLogger(__name__)

MULTI_WS_REGEX = re.compile(r"\n\n+")

ZERO_TIMEDELTA = timedelta(0)

# Info message if truthy return -> Function taking a Subtitle, skip if True
SUBTITLE_SKIP_CONDITIONS = (
    ("No content", lambda sub: not sub.content.strip()),
    ("Start time < 0 seconds", lambda sub: sub.start < ZERO_TIMEDELTA),
    ("Subtitle start time >= end time", lambda sub: sub.start >= sub.end),
)

SECONDS_IN_HOUR = 3600
SECONDS_IN_MINUTE = 60
HOURS_IN_DAY = 24
MICROSECONDS_IN_MILLISECOND = 1000


@functools.total_ordering
class Subtitle:
    r"""
    The metadata relating to a single subtitle. Subtitles are sorted by start
    time by default. If no index was provided, index 0 will be used on writing
    an SRT block.

    :param index: The SRT index for this subtitle
    :type index: int or None
    :param start: The time that the subtitle should start being shown
    :type start: :py:class:`datetime.timedelta`
    :param end: The time that the subtitle should stop being shown
    :type end: :py:class:`datetime.timedelta`
    :param str content: The subtitle content. Should not contain OS-specific
                        line separators, only \\n. This is taken care of
                        already if you use :py:func:`srt.parse` to generate
                        Subtitle objects.
    """

    # pylint: disable=R0913
    def __init__(
        self, index: Union[int, None], start: timedelta, end: timedelta, content: str
    ) -> None:
        self.index = index
        self.start = start
        self.end = end
        self.content = content

    def __hash__(self) -> int:
        return hash(frozenset(vars(self).items()))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Subtitle):
            return NotImplemented

        return vars(self) == vars(other)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Subtitle):
            return NotImplemented

        return (self.start, self.end, self.index) < (
            other.start,
            other.end,
            other.index,
        )

    def __repr__(self) -> str:
        # Python 2/3 cross compatibility
        var_items = getattr(vars(self), "iteritems", getattr(vars(self), "items"))
        item_list = ", ".join(f"{k}={v!r}" for k, v in var_items())
        return f"{type(self).__name__}({item_list})"

    def to_srt(self, eol: Union[str, None] = None) -> str:
        r"""
        Convert the current :py:class:`Subtitle` to an SRT block.

        :param str eol: The end of line string to use (default "\\n")
        :returns: The metadata of the current :py:class:`Subtitle` object as an
                  SRT formatted subtitle block
        :rtype: str
        """
        output_content = make_legal_content(self.content)

        if eol is None:
            eol = "\n"
        elif eol != "\n":
            output_content = output_content.replace("\n", eol)

        template = "{idx}{eol}{start} --> {end}{eol}{content}{eol}{eol}"
        return template.format(
            idx=self.index or 0,
            start=timedelta_to_srt_timestamp(self.start),
            end=timedelta_to_srt_timestamp(self.end),
            content=output_content,
            eol=eol,
        )


def make_legal_content(content: str) -> str:
    r"""
    Remove illegal content from a content block. Illegal content includes:

    * Blank lines
    * Starting or ending with a blank line

    .. doctest::

        >>> make_legal_content('\nfoo\n\nbar\n')
        'foo\nbar'

    :param str content: The content to make legal
    :returns: The legalised content
    :rtype: srt
    """
    # Optimisation: Usually the content we get is legally valid. Do a quick
    # check to see if we really need to do anything here. This saves time from
    # generating legal_content by about 50%.
    if content and content[0] != "\n" and "\n\n" not in content:
        return content

    legal_content = MULTI_WS_REGEX.sub("\n", content.strip("\n"))
    LOG.info("Legalised content %r to %r", content, legal_content)
    return legal_content


def timedelta_to_srt_timestamp(timedelta_timestamp: timedelta) -> str:
    r"""
    Convert a :py:class:`~datetime.timedelta` to an SRT timestamp.

    .. doctest::

        >>> import datetime
        >>> delta = datetime.timedelta(hours=1, minutes=23, seconds=4)
        >>> timedelta_to_srt_timestamp(delta)
        '01:23:04,000'

    :param datetime.timedelta timedelta_timestamp: A datetime to convert to an
                                                   SRT timestamp
    :returns: The timestamp in SRT format
    :rtype: str
    """

    hrs, secs_remainder = divmod(timedelta_timestamp.seconds, SECONDS_IN_HOUR)
    hrs += timedelta_timestamp.days * HOURS_IN_DAY
    mins, secs = divmod(secs_remainder, SECONDS_IN_MINUTE)
    msecs = timedelta_timestamp.microseconds // MICROSECONDS_IN_MILLISECOND
    return f"{int(hrs):02}:{int(mins):02}:{int(secs):02},{int(msecs):03}"


def sort_and_reindex(
    subtitles: Union[Generator[Subtitle, None, None], List[Subtitle]],
    start_index: int = 1,
    in_place: bool = False,
    skip: bool = True,
) -> Generator[Subtitle, None, None]:
    """
    Reorder subtitles to be sorted by start time order, and rewrite the indexes
    to be in that same order. This ensures that the SRT file will play in an
    expected fashion after, for example, times were changed in some subtitles
    and they may need to be resorted.

    If skip=True, subtitles will also be skipped if they are considered not to
    be useful. Currently, the conditions to be considered "not useful" are as
    follows:

    - Content is empty, or only whitespace
    - The start time is negative
    - The start time is equal to or later than the end time

    .. doctest::

        >>> from datetime import timedelta
        >>> one = timedelta(seconds=1)
        >>> two = timedelta(seconds=2)
        >>> three = timedelta(seconds=3)
        >>> subs = [
        ...     Subtitle(index=999, start=one, end=two, content='1'),
        ...     Subtitle(index=0, start=two, end=three, content='2'),
        ... ]
        >>> list(sort_and_reindex(subs))  # doctest: +ELLIPSIS
        [Subtitle(...index=1...), Subtitle(...index=2...)]

    :param subtitles: :py:class:`Subtitle` objects in any order
    :param int start_index: The index to start from
    :param bool in_place: Whether to modify subs in-place for performance
                          (version <=1.0.0 behaviour)
    :param bool skip: Whether to skip subtitles considered not useful (see
                      above for rules)
    :returns: The sorted subtitles
    :rtype: :term:`generator` of :py:class:`Subtitle` objects
    """
    skipped_subs = 0
    for sub_num, subtitle in enumerate(sorted(subtitles), start=start_index):
        if not in_place:
            subtitle = Subtitle(**vars(subtitle))

        if skip:
            try:
                _should_skip_sub(subtitle)
            except _ShouldSkipException as thrown_exc:
                if subtitle.index is None:
                    LOG.info("Skipped subtitle with no index: %s", thrown_exc)
                else:
                    LOG.info(
                        "Skipped subtitle at index %d: %s", subtitle.index, thrown_exc
                    )
                skipped_subs += 1
                continue

        subtitle.index = sub_num - skipped_subs

        yield subtitle


def _should_skip_sub(subtitle: Subtitle) -> None:
    """
    Check if a subtitle should be skipped based on the rules in
    SUBTITLE_SKIP_CONDITIONS.

    :param subtitle: A :py:class:`Subtitle` to check whether to skip
    :raises _ShouldSkipException: If the subtitle should be skipped
    """
    for info_msg, sub_skipper in SUBTITLE_SKIP_CONDITIONS:
        if sub_skipper(subtitle):
            raise _ShouldSkipException(info_msg)


def compose(
    subtitles: Union[Generator[Subtitle, None, None], List[Subtitle]],
    reindex: bool = True,
    start_index: int = 1,
    eol: Union[str, None] = None,
    in_place: bool = False,
) -> str:
    r"""
    Convert an iterator of :py:class:`Subtitle` objects to a string of joined
    SRT blocks.

    .. doctest::

        >>> from datetime import timedelta
        >>> start = timedelta(seconds=1)
        >>> end = timedelta(seconds=2)
        >>> subs = [
        ...     Subtitle(index=1, start=start, end=end, content='x'),
        ...     Subtitle(index=2, start=start, end=end, content='y'),
        ... ]
        >>> compose(subs)  # doctest: +ELLIPSIS
        '1\n00:00:01,000 --> 00:00:02,000\nx\n\n2\n00:00:01,000 --> ...'

    :param subtitles: The subtitles to convert to SRT blocks
    :type subtitles: :term:`iterator` of :py:class:`Subtitle` objects
    :param bool reindex: Whether to reindex subtitles based on start time
    :param int start_index: If reindexing, the index to start reindexing from
    :param str eol: The end of line string to use (default "\\n")
    :returns: A single SRT formatted string, with each input
              :py:class:`Subtitle` represented as an SRT block
    :param bool in_place: Whether to reindex subs in-place for performance
                          (version <=1.0.0 behaviour)
    :rtype: str
    """
    if reindex:
        subtitles = sort_and_reindex(
            subtitles, start_index=start_index, in_place=in_place
        )

    return "".join(subtitle.to_srt(eol=eol) for subtitle in subtitles)


class _ShouldSkipException(Exception):
    """
    Raised when a subtitle should be skipped.
    """



"""Functions to play audio on Windows using native win32 APIs"""

import sys

from .util import pr_err


def play_mp3_win32(mp3_fname: str) -> None:
    """Play mp3 file with given path using win32 API"""

    if sys.platform != "win32":
        raise NotImplementedError("Function only available on Windows")

    # pylint: disable-next=import-outside-toplevel
    from ctypes import create_unicode_buffer, windll, wintypes  # type: ignore

    _get_short_path_name_w = windll.kernel32.GetShortPathNameW
    _get_short_path_name_w.argtypes = [
        wintypes.LPCWSTR,
        wintypes.LPWSTR,
        wintypes.DWORD,
    ]
    _get_short_path_name_w.restype = wintypes.DWORD

    def get_short_path_name(long_name: str) -> str:
        """
        Gets the DOS-safe short path name of a given long path.
        http://stackoverflow.com/a/23598461/200291
        """
        output_buf_size = 0
        while True:
            output_buf = create_unicode_buffer(output_buf_size)
            needed = _get_short_path_name_w(long_name, output_buf, output_buf_size)
            if output_buf_size >= needed:
                return output_buf.value
            output_buf_size = needed

    mci_send_string_w = windll.winmm.mciSendStringW

    def mci_send(msg: str) -> None:
        """Send MCI command string"""
        result = mci_send_string_w(msg, 0, 0, 0)
        if result != 0:
            pr_err(f"Error {result} in mciSendString {msg}. Exiting.")
            sys.exit(1)

    mp3_shortname = get_short_path_name(mp3_fname)

    mci_send("Close All")
    mci_send(f'Open "{mp3_shortname}" Type MPEGVideo Alias theMP3')
    mci_send("Play theMP3 Wait")
    mci_send("Close theMP3")



"""Main entrypoint for the edge-playback package."""

import argparse
import os
import subprocess
import sys
import tempfile
from shutil import which
from typing import List, Optional, Tuple

from .util import pr_err


def _parse_args() -> Tuple[bool, List[str]]:
    parser = argparse.ArgumentParser(
        prog="edge-playback",
        description="Speak text using Microsoft Edge's online text-to-speech API",
        epilog="See `edge-tts` for additional arguments",
    )
    parser.add_argument(
        "--mpv",
        action="store_true",
        help="Use mpv to play audio. By default, false on Windows and true on all other platforms",
    )
    args, tts_args = parser.parse_known_args()
    use_mpv = sys.platform != "win32" or args.mpv
    return use_mpv, tts_args


def _check_deps(use_mpv: bool) -> None:
    depcheck_failed = False
    deps = ["edge-tts"]
    if use_mpv:
        deps.append("mpv")

    for dep in deps:
        if not which(dep):
            pr_err(f"{dep} is not installed.")
            depcheck_failed = True

    if depcheck_failed:
        pr_err("Please install the missing dependencies.")
        sys.exit(1)


def _create_temp_files(
    use_mpv: bool, mp3_fname: Optional[str], srt_fname: Optional[str], debug: bool
) -> Tuple[str, Optional[str]]:
    media = subtitle = None
    if not mp3_fname:
        media = tempfile.NamedTemporaryFile(suffix=".mp3", delete=False)
        media.close()
        mp3_fname = media.name
        if debug:
            print(f"Media file: {mp3_fname}")

    if not srt_fname and use_mpv:
        subtitle = tempfile.NamedTemporaryFile(suffix=".srt", delete=False)
        subtitle.close()
        srt_fname = subtitle.name

    if debug and srt_fname:
        print(f"Subtitle file: {srt_fname}\n")

    return mp3_fname, srt_fname


def _run_edge_tts(
    mp3_fname: str, srt_fname: Optional[str], tts_args: List[str]
) -> None:
    edge_tts_cmd = ["edge-tts", f"--write-media={mp3_fname}"]
    if srt_fname:
        edge_tts_cmd.append(f"--write-subtitles={srt_fname}")
    edge_tts_cmd = edge_tts_cmd + tts_args
    with subprocess.Popen(edge_tts_cmd) as process:
        process.communicate()


def _play_media(use_mpv: bool, mp3_fname: str, srt_fname: Optional[str]) -> None:
    if sys.platform == "win32" and not use_mpv:
        # pylint: disable-next=import-outside-toplevel
        from .win32_playback import play_mp3_win32

        play_mp3_win32(mp3_fname)
        return

    mpv_cmd = [
        "mpv",
        "--msg-level=all=error,statusline=status",
    ]
    if srt_fname:
        mpv_cmd.append(f"--sub-file={srt_fname}")
    mpv_cmd.append(mp3_fname)
    with subprocess.Popen(mpv_cmd) as process:
        process.communicate()


def _cleanup(mp3_fname: Optional[str], srt_fname: Optional[str], keep: bool) -> None:
    if keep and mp3_fname is not None:
        print(f"\nKeeping temporary files: {mp3_fname}", end="")
        if srt_fname:
            print(f" and {srt_fname}", end="")
        print()
        return

    if mp3_fname is not None and os.path.exists(mp3_fname):
        os.unlink(mp3_fname)
    if srt_fname is not None and os.path.exists(srt_fname):
        os.unlink(srt_fname)


def _main() -> None:
    use_mpv, tts_args = _parse_args()
    _check_deps(use_mpv)

    debug = os.environ.get("EDGE_PLAYBACK_DEBUG") is not None
    keep = os.environ.get("EDGE_PLAYBACK_KEEP_TEMP") is not None
    mp3_fname = os.environ.get("EDGE_PLAYBACK_MP3_FILE")
    srt_fname = os.environ.get("EDGE_PLAYBACK_SRT_FILE")

    try:
        mp3_fname, srt_fname = _create_temp_files(use_mpv, mp3_fname, srt_fname, debug)
        _run_edge_tts(mp3_fname, srt_fname, tts_args)
        _play_media(use_mpv, mp3_fname, srt_fname)
    finally:
        _cleanup(mp3_fname, srt_fname, keep)


if __name__ == "__main__":
    _main()



from setuptools import setup

# Metadata goes in setup.cfg. These are here for GitHub's dependency graph.
setup(
    name="edge-tts",
    install_requires=[
        "aiohttp>=3.8.0,<4.0.0",
        "certifi>=2023.11.17",
        "tabulate>=0.4.4,<1.0.0",
        "typing-extensions>=4.1.0,<5.0.0",
    ],
)
